import Foundation


/**
Verwaltet ein Knotensymbol in einem Graph

- author: Albert Wiedemann
- version: 1.0
*/
class KnotenSymbol
{
    /** Außenkreis des Darstellungssymbols */
    private let außen: Kreis
    /** Innenkreis des Darstellungssymbols */
    private let innen: Kreis
    /** Text des Darstellungssymbols */
    private let text: Text
    /** Der Bezeichner */
    private var bezeichner: String
    /** Radius */
    private var r: Int
    /** x-Koordinate */
    private var x: Int
    /** y-Koordinate */
    private var y: Int
    /** Farbe */
    private var farbe: String
    
    /**
    Legt das Symbol an und besetzt die Attribute.
    - parameters:
        - x: x-Koordinate
        - y: y-Koordinate
        - radius: Radius
        - farbe: Farbe
        - bezeichner: Bezeichner
    */
    init(x: Int, y: Int, radius: Int, farbe: String, bezeichner: String)
    {
        self.x = x
        self.y = y
        self.r = radius
        self.farbe = farbe
        self.bezeichner = bezeichner
        außen = Kreis()
        außen.FarbeSetzen(farbe: "schwarz")
        innen = Kreis()
        text = Text()
        text.TextGrößeSetzen(größe: 18)
        DarstellungAktualisieren()
    }
    
    /**
    Aktualisiert das Symbol
    */
    private func DarstellungAktualisieren()
    {
        außen.PositionSetzen(x: x, y: y)
        außen.RadiusSetzen(radius: r)
        innen.PositionSetzen(x: x, y: y)
        innen.RadiusSetzen(radius: r - 2)
        innen.FarbeSetzen(farbe: farbe)
        text.PositionSetzen(x: x - bezeichner.count / 2 * 10 - bezeichner.count % 2 * 5, y: y + 6)
        text.TextSetzen(text: bezeichner)
    }
    
    /**
    Setzt die Farbe der Darstellung
    - parameters:
        - f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        farbe = f
        DarstellungAktualisieren()
    }
    
    /**
    Setzt die Position der Darstellung
    - parameters:
        - x: x-Koordinate
        - y: y-Koordinate
    */
    func PositionSetzen(x: Int, y: Int)
    {
        self.x = x
        self.y = y
        DarstellungAktualisieren()
    }
    
    /**
    Setzt den Bezeichner des Symbols
    - parameters:
        - bezeichner: der (neue) Bezeichner
    */
    func BezeichnerSetzen(bezeichner: String)
    {
        self.bezeichner = bezeichner
        DarstellungAktualisieren()
    }
    
    /**
    Meldet die x-Koordinate des Symbols zurück.
    - returns: x-Koordinate
    */
    func XGeben() -> Int
    {
        return x
    }
    
    /**
    Meldet die y-Koordinate des Symbols zurück.
    - returns: y-Koordinate
    */
    func YGeben() -> Int
    {
        return y
    }
    
    /**
    Meldet den Bezeichner des Symbols zurück.
    - returns: Bezeichner
     */
    func BezeichnerGeben() -> String
    {
        return bezeichner
    }
    
    /**
    Entfernt das Knotensymbol aus der Anzeige
    */
    func Entfernen()
    {
        außen.Entfernen()
        innen.Entfernen()
        text.Entfernen()
    }
    
    /**
    Setzt die Sichtbarkeit des Symbols
    - parameters:
        - sichtbar: wenn wahr, ist das Symbol sichtbar
    */
    func SichtbarkeitSetzen(sichtbar: Bool)
    {
        außen.SichtbarkeitSetzen(sichtbar: sichtbar)
        innen.SichtbarkeitSetzen(sichtbar: sichtbar)
        text.SichtbarkeitSetzen(sichtbar: sichtbar)
    }
}


/**
Verwaltet ein Kantensymbol in einem Graph

- author: Albert Wiedemann
- version: 1.0
*/
class KantenSymbol
{
    /** Außenrechteck des Darstellungssymbols */
    private let außen: Rechteck
    /** Innenrechteck des Darstellungssymbols */
    private let innen: Rechteck
    /** Richtungspfeil des Darstellungssymbols */
    private let pfeil: Dreieck
    /** Gewichtsangabe des Darstellungssymbols */
    private let text: Text
    /** Breite des Darstellungssymbols */
    private let breite: Int
    /** Gerichtete Kante */
    private let gerichtet: Bool
    /** Gewicht der Kante */
    private let gewicht: String
    /** Farbe */
    private var farbe: String
    /** Startknoten */
    private let von: KnotenSymbol
    /** Endknoten */
    private let nach: KnotenSymbol

    /**
    Legt das Symbol an und besetzt die Attribute.
    - parameters:
        - start: Startknoten
        - ende: Endknoten
        - gerichtet: wenn wahr, ist der Weg gerichtet
        - gewicht: Kantengewicht
        - breite: Breite der Linie
        - farbe: Farbe der Linie
    */
    init(start: KnotenSymbol, ende: KnotenSymbol, gerichtet: Bool, gewicht: String, breite: Int, farbe: String)
    {
        von = start
        nach = ende;
        self.farbe = farbe
        self.breite = breite
        self.gerichtet = gerichtet
        self.gewicht = gewicht
        außen = Rechteck()
        innen = Rechteck()
        pfeil = Dreieck()
        pfeil.FarbeSetzen(farbe: "schwarz")
        text = Text()
        text.SichtbarkeitSetzen(sichtbar: gewicht != "")
        text.TextSetzen(text: gewicht)
        DarstellungAktualisieren()
    }
    
    /**
    Aktualisiert das Symbol
    */
    func DarstellungAktualisieren()
    {
        var x1 = Double(von.XGeben())
        var y1 = Double(von.YGeben())
        var x2 = Double(nach.XGeben())
        var y2 = Double(nach.YGeben())
        let länge = Int(sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)))
        let xm = Int((x1 + x2) / 2.0)
        let ym = Int((y1 + y2) / 2.0)
        let x = xm - länge / 2
        let y = ym - breite / 2
        außen.PositionSetzen(x: x, y: y)
        außen.GrößeSetzen(breite: länge, höhe: breite)
        if breite < 4
        {
            let größe = breite * 4;
            innen.SichtbarkeitSetzen(sichtbar: false)
            außen.FarbeSetzen(farbe: farbe)
            pfeil.SichtbarkeitSetzen(sichtbar: gerichtet)
            pfeil.PositionSetzen(x: xm, y: ym - größe / 2)
            pfeil.GrößeSetzen(breite: größe, höhe: größe)
            pfeil.WinkelSetzen(winkel: 270)
        }
        else
        {
            let delta = breite <= 8 ? 2 : 4
            innen.SichtbarkeitSetzen(sichtbar: true)
            innen.PositionSetzen(x: x, y: y + delta / 2)
            innen.GrößeSetzen(breite: länge, höhe: breite - delta)
            innen.FarbeSetzen(farbe: farbe)
            außen.FarbeSetzen(farbe: "schwarz")
            pfeil.SichtbarkeitSetzen(sichtbar: gerichtet)
            pfeil.PositionSetzen(x: xm, y: ym - breite / 2)
            pfeil.GrößeSetzen(breite: breite - delta, höhe: breite)
            pfeil.WinkelSetzen(winkel: 270)
        }
        
        var winkel = 0.0
        if x1 == x2
        {
            if y2 > y1
            {
                winkel = -90
            }
            else
            {
                winkel = 90
            }
        }
        else if x1 < x2
        {
            winkel = -atan((y2 - y1) / (x2 - x1)) / Double.pi * 180.0
        }
        else
        {
            winkel = 180 - atan((y2 - y1) / (x2 - x1)) / Double.pi * 180.0
        }
        pfeil.WinkelSetzen(winkel: Int(winkel + 270))
        innen.WinkelSetzen(winkel: Int(winkel))
        außen.WinkelSetzen(winkel: Int(winkel))
        pfeil.GanzNachHintenBringen()
        innen.GanzNachHintenBringen()
        außen.GanzNachHintenBringen()
        let delta = Double(breite)
        let w = (winkel + 90) / 180.0 * Double.pi
        if (-90 <= winkel) && (winkel <= 0)
        {
            text.PositionSetzen(x: xm + Int(delta * cos(w)), y: ym - Int(delta * sin(w)))
        }
        else if (0 < winkel) && (winkel <= 90)
        {
            text.PositionSetzen(x: xm - Int(delta * cos(w)), y: ym + Int(delta * sin(w)) + 7)
        }
        else if (90 < winkel) && (winkel <= 180)
        {
            text.PositionSetzen(x: xm - Int(delta * cos(w)), y: ym + Int(delta * sin(w)))
        }
        else
        {
            text.PositionSetzen(x: xm + Int(delta * cos(w)), y: ym - Int(delta * sin(w)) + 10)
        }
    }
    
    /**
    Entfernt das Kantensymbol aus der Anzeige
    */
    func Entfernen()
    {
        außen.Entfernen()
        innen.Entfernen()
        pfeil.Entfernen()
        text.Entfernen()
    }
    
    /**
    Setzt die Farbe der Darstellung
    - parameters:
        - f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        farbe = f
        DarstellungAktualisieren()
    }
    
    /**
    Meldet den Startknoten
    - returns: Startknoten
    */
    func StartsymbolGeben() -> KnotenSymbol
    {
        return von
    }
    
    /**
    Meldet den Zielknoten
    - returns: Zielknoten
    */
    func ZielsymbolGeben() -> KnotenSymbol
    {
        return nach
    }
}

/**
Die Klasse Wörterbucheintrag dient zur Speicherung von
Wörterbucheinträgen für ein Englisch-Deutsches
Wörterbuch. Dabei kann jedem englischen Wort mehrere
deutsche Bedeutungen zugeordnet werden, wobei alle Bedeutungen in
einer einzigen Zeichenkette Bedeutung gespeichert werden.

- author: Peter Brichzin
- version: 23.05.24
*/
class Wörterbucheintrag
{
    /** Englisches Wort;  Schlüssel dieses Datenelements */
    private var wort: String
    /** Deutsche Bedeutung; mehrere Bedeutungen werden als eine Zeichenkette zusammengefasst */
    private var bedeutung: String

    /**
    Konstruktor für Objekte der Klasse Wörterbucheintrag
    - parameters:
        - wortNeu: das englische Wort.
        - bedeutungNeu die deutsche(n) Bedeutung(en).
    */
    init(wortNeu: String, bedeutungNeu: String)
    {
        wort = wortNeu
        bedeutung = bedeutungNeu
    }

    /**
    Gibt Information über das Datenelement zu Kontrollzwecken
    auf das Terminalfenster aus.
    */
    func InformationAusgeben()
    {
        print(wort + ": " + bedeutung)
    }

    /**
    Vergleicht zwei Datenelemente auf Gleichheit.
    - parameters:
        - dvergleich: Datenelement mit dem das Objekt verglichen wird.
    - returns: true, wenn die beiden Datenelemente gleichen Schlüssel haben.
    */
    func IstGleich(dvergleich: Wörterbucheintrag) -> Bool
    {
        if wort == dvergleich.WortGeben()
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Vergleicht zwei Datenelemente bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichselement: Datenelement mit dem das Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen größeren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func IstGrößerAls(dvergleich: Wörterbucheintrag) -> Bool
    {
        if wort > dvergleich.WortGeben()
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, auf Gleichheit.
    - parameters:
        - vergleichsSchlüssel: Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn die beiden Schlüssel gleich sind.
    */
    func SchlüsselIstGleich(vergleichsSchlüssel: String) -> Bool
    {
        return wort == vergleichsSchlüssel
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichsSchlüssel Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen größeren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func SchlüsselIstGrößerAls(vergleichsSchlüssel: String) -> Bool
    {
        if wort > vergleichsSchlüssel
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Vergleicht zwei Schlüssel, die als Zeichenketten vorliegen, bezüglich der Ordnungsrelation.
    - parameters:
        - vergleichsSchüssel Schlüssel  mit dem der Schlüssel des Objekt verglichen wird.
    - returns: true, wenn das aktuelle Element einen kleineren Schlüssel hat, als das angegebene Vergleichselement.
    */
    func SchlüsselIstKleinerAls(vergleichsSchlüssel: String) -> Bool
    {
        if wort < vergleichsSchlüssel
        {
            return true
        }
        else
        {
            return false
        }
    }

    /**
    Gibt den Schlüssel des Datenelements als String aus, auch wenn der Datentyp vom
    String abweicht. Diese Methode wird beim Entfernen eines Knotens benötigt.
    - returns: Schlüssel als String.
     */
    func SchlüsselAlsStringGeben() -> String
    {
        return wort
    }

    /**
    Geben-Methode zum Attribut wort
    - returns: das englische Wort
    */
    func WortGeben() -> String
    {
        return wort
    }

    /**
    Geben-Methode zum Attribut bedeutung
    - returns: bedeutung: die deutsche(n) Bedeutung(en) als eine Zeichenkette
    */
    func BedeutungGeben() -> String
    {
        return bedeutung
    }

    /**
    Setzen-Methode zum Attribut bedeutung
    - parameters:
        - bedeutungNeu: neue Bedeutung für das englische Wort
    */
    func BedeutungSetzen(bedeutungNeu: String)
    {
        bedeutung = bedeutungNeu
    }
}

/**
Abstrakte Klasse Baumelement zur Realisierung einer Baumstruktur mit Hilfe des Entwurfsmusters
Kompositum: Die Klasse legt ausschließlich die Schnittstelle für die Klassen
Knoten und Abschluss in Form von Methoden fest.

- author: Peter Brichzin
- version: 23.5.24
*/
class Baumelement
{
    /**
     Der Konstruktor ist nur der Vollständigkeit halber angegeben.
     */
    init()
    {
        // nichts zu tun
    }
    
    /**
     Fügt ein Datenelement, falls sein Schlüssel noch nicht vorhanden ist,
     in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum
     bleibt erhalten.
     - parameters:
        - datenNeu: das Datenelement des einzufügenden Knoten
     - returns: (neues) Nachfolgeelement des Aufrufers
     */
    func Einfügen(datenNeu: Wörterbucheintrag) -> Baumelement
    {
        fatalError("Baumelement.Einfügen muss überschrieben werden.")
    }
    
    /**
     Sucht ein Datenelement, das über seinen Schlüssel identifiziert wird
     und gibt dieses aus.
     - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
     - returns: gesuchtes Datenelement bzw. nil, falls die Suche erfolglos ist
     */
    func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        fatalError("Baumelement.Suchen muss überschrieben werden.")
    }
    
    /**
     Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
     Zeichenkette) vorhanden ist.
     - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
     - returns: true im Erfolgsfall, false sonst.
     */
    func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        fatalError("Baumelement.IstVorhanden muss überschrieben werden.")
    }
    
    /**
     Berechnet die Höhe des Baums und gibt diesen Wert zurück
     - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
     - returns: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
     */
    func HöheGeben(aktTiefe: Int) -> Int
    {
        fatalError("Baumelement.HöheGeben muss überschrieben werden.")
    }
    
    /**
     Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
     - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe: die Tiefe des aktuellen Knotens
     - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
     */
    func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        fatalError("Baumelement.TiefeGeben muss überschrieben werden.")
    }
    
    /**
    Zeichnet den Knoten passend im Zeichenfenster, ergänzt ggf. noch eine Kante zum Vorgänger
    - parameters:
        - xmin: der minimale Wert des Breitenbereichs
        - xmax: der maximale Wert des Breitenbereichs
        - y: der y-Wert der Ebene
        - dy: der Unerschied zur nächsten Ebene
        - symbolVorgänger: nötig, um die Kante zu zeichnen: das eigene Symbol -> ist für den neuen Knoten das Vorgängersymbol
     */
    func KnotenZeichnen(xmin: Int, xmax: Int, y: Int, dy: Int, symbolVorgänger: KnotenSymbol?)
    {
        fatalError("Baumelement.KnotenZeichnen muss überschrieben werden.")
    }
}

/**
Klasse Abschluss als Bestandteil eines geordneten Binärbaums.
(Implementierung mit dem Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
*/
class Abschluss: Baumelement
{
    /** Das Darstellungssymbol des Knotens  */
    private var knotenSymbol: KnotenSymbol
    /** Das Darstellungssymbol  der Kanten */
    private var kantenSymbol: KantenSymbol?

    /**
    Konstruktor für Objekte der Klasse Abschluss
    */
    override init()
    {
        knotenSymbol = KnotenSymbol(x: 100, y: 100, radius: 25, farbe: "magenta", bezeichner: "a")
        knotenSymbol.SichtbarkeitSetzen(sichtbar: true)
        kantenSymbol = nil
        super.init()
    }

    /**
    Fügt ein Datenelement in die bestehende Datenstruktur ein, indem es einen neuen Knoten
    (mit zwei Abschluss-Objekten als Nachfolger) erzeugt und eine Referenz auf diesen Knoten
    zurückgibt. Das Abschluss-Objekt, dass diese Methode ausführt, wird dabei aus dem Baum entfernt.
    - parameters:
        - datenNeu: das Datenelement des einzufügenden Knoten
    - returns: Referenz auf den neu erzeugten Knoten
    */
    override func Einfügen(datenNeu: Wörterbucheintrag) -> Baumelement
    {
        // Dieses Abschlussobjekt ist am Ende des Methodenaufrufs nicht mehr Teil der
        // Struktur, deshalb muss das KnotenSymbol ggf. auch das Kantensymbol entfernt werden.
        knotenSymbol.Entfernen()
        if kantenSymbol != nil
        {
            kantenSymbol!.Entfernen()
        }
        
        return Knoten(datenNeu: datenNeu)
    }

    /**
    Sucht ein Datenelement, das über seinen Schlüssel identifiziert wird
    und gibt dieses aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
    - returns: nil, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
    */
    override func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        return nil
    }

    /**
    Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: false, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
    */
    override func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        return false
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die übergebene Höhe -1, da hier beim Abschluss das Ende eines Astes erreicht ist und die Abschlusselemente bei der Höhenberechnung nicht einbezogen werden.
    */
    override func HöheGeben(aktTiefe: Int) -> Int
    {
        return aktTiefe - 1
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe: die Tiefe des aktuellen Knotens
    - returns: -1 da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
    */
    override func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        return -1
    }
    
    /**
    Zeichnet den Knoten passend im Zeichenfenster, ergänzt ggf. noch eine Kante zum Vorgänger
    - parameters:
        - xmin: der minimale Wert des Breitenbereichs
        - xmax: der maximale Wert des Breitenbereichs
        - y: der y-Wert der Ebene
        - dy: der Unerschied zur nächsten Ebene
        - symbolVorgänger: nötig, um die Kante zu zeichnen: das eigene Symbol -> ist für den neuen Knoten das Vorgängersymbol
     */
    override func KnotenZeichnen(xmin: Int, xmax: Int, y: Int, dy: Int, symbolVorgänger: KnotenSymbol?)
    {
        knotenSymbol.PositionSetzen(x: xmin + (xmax - xmin) / 2, y: y)

        // Eine Kante wird nur gezeichnet, wenn es einen Vorgänger gibt (d.h. der Baum nicht leer ist)
        if symbolVorgänger != nil
        {
            if kantenSymbol == nil
            {
                kantenSymbol = KantenSymbol(start: symbolVorgänger!, ende: knotenSymbol, gerichtet: false, gewicht: "", breite: 3, farbe: "schwarz")
            }
            else
            {
                kantenSymbol!.DarstellungAktualisieren()
            }
        }
    }
}

/**
Klasse Knoten als Bestandteil eines geordneten Binärbaums.
(Implementierung mit dem Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 23.5.24
*/
class Knoten: Baumelement
{

    /** Datenelement des Knotens */
    private var daten: Wörterbucheintrag
    /** Referenz auf den linken Nachfolger */
    private var linkerNachfolger: Baumelement
    /** Referenz auf den rechten Nachfolger */
    private var rechterNachfolger: Baumelement

    /** Das Darstellungssymbol des Knotens  */
    private var knotenSymbol: KnotenSymbol
    /** Das Darstellungssymbol  der Kanten */
    private var kantenSymbol: KantenSymbol?

    /**
     * Konstruktor für Objekte der Klasse Knoten ohne linken und rechten Nachfolger, d.h. es werden
     * zwei Abschluss-Objekte erzeugt und den entsprechenden Referenzattributen zugewiesen.
     * @param datenNeu Referenz auf ein Objekt der KLasse Wörterbucheintrag, das vom Knoten verwaltet wird.
     */
    init(datenNeu: Wörterbucheintrag)
    {
        daten = datenNeu;
        linkerNachfolger = Abschluss()
        rechterNachfolger = Abschluss()
        knotenSymbol = KnotenSymbol(x: 100, y: 100, radius: 25, farbe: "cyan", bezeichner: daten.SchlüsselAlsStringGeben())
        knotenSymbol.SichtbarkeitSetzen(sichtbar: true)
        kantenSymbol = nil
        super.init()
    }

    /**
    Fügt ein Datenelement (Wörterbucheintrag), falls sein Schlüssel noch nicht vorhanden ist,
    in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum
    bleibt erhalten.
    - parameters:
        - datenNeu: einzufügendes Datenelement (Wörterbucheintrag)
    - returns: Referenz auf das eigene Objekt
    */
    override func Einfügen(datenNeu: Wörterbucheintrag) -> Baumelement
    {
        // Falls es den Wert schon gibt, Einfügen abbrechen.
        if daten.IstGleich(dvergleich: datenNeu)
        {
            print("Die Daten existieren schon!")
        }
        // Abfrage ob im rechten oder linken Teilbaum eingefügt werden muss
        else
        {
            if daten.IstGrößerAls(dvergleich: datenNeu)
            {
                linkerNachfolger = linkerNachfolger.Einfügen(datenNeu: datenNeu)
            }
            else
            {
                rechterNachfolger = rechterNachfolger.Einfügen(datenNeu: datenNeu)
            }
        }
        return self
    }

    /**
    Sucht ein Datenelement (Wörterbucheintrag), das über seinen Schlüssel identifiziert wird
    und gibt dieses aus.
    - parameters:
        - suchSchlüssel: Schlüssel nach dem gesucht wird als Zeichenkette
    - returns: gesuchtes Datenelement (Wörterbucheintrag) bzw. nil, falls die Suche erfolglos ist
    */
    override func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return daten
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.Suchen(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
            else
            {
                return rechterNachfolger.Suchen(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
        }
    }

    /**
    Überprüft, ob ein Datenelement (Wörterbucheintrag) mit dem eingegebenen Schlüssel (in Form einer
    Zeichenkette) vorhanden ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
    - returns: true im Erfolgsfall, false sonst.
    */
    override func IstVorhanden(suchSchlüssel: String) -> Bool
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return true
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.IstVorhanden(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
            else
            {
                return rechterNachfolger.IstVorhanden(suchSchlüssel: suchSchlüssel)  // Rückgabewert wird nach unten abgefragt und dann nach oben durchgereicht
            }
        }
    }

    /**
    Berechnet die Höhe des Baums und gibt diesen Wert zurück
    - parameters:
        - aktTiefe: Tiefe des aktuellen Knotens
    - returns: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
    */
    override func HöheGeben(aktTiefe: Int) -> Int
    {
        // lokale Attriubute
        let maxHöheLinkerTeilbaum = linkerNachfolger.HöheGeben(aktTiefe: aktTiefe + 1)
        let maxHöheRechterTeilbaum = rechterNachfolger.HöheGeben(aktTiefe: aktTiefe + 1)

        if maxHöheLinkerTeilbaum > maxHöheRechterTeilbaum
        {
            return maxHöheLinkerTeilbaum
        }
        else
        {
            return maxHöheRechterTeilbaum
        }
    }

    /**
    Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
    - parameters:
        - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
        - tiefe: die Tiefe des aktuellen Knotens
    - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
    */
    override func TiefeGeben(suchSchlüssel: String, tiefe: Int) -> Int
    {
        if daten.SchlüsselIstGleich(vergleichsSchlüssel: suchSchlüssel)
        {
            return tiefe
        }
        else
        {
            if daten.SchlüsselIstGrößerAls(vergleichsSchlüssel: suchSchlüssel)
            {
                return linkerNachfolger.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: tiefe + 1)
            }
            else
            {
                return rechterNachfolger.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: tiefe + 1)
            }
        }
    }

    /**
    Zeichnet den Knoten passend im Zeichenfenster, ergänzt ggf. noch eine Kante zum Vorgänger
    - parameters:
        - xmin: der minimale Wert des Breitenbereichs
        - xmax: der maximale Wert des Breitenbereichs
        - y: der y-Wert der Ebene
        - dy: der Unerschied zur nächsten Ebene
        - symbolVorgänger: nötig, um die Kante zu zeichnen: das eigene Symbol -> ist für den neuen Knoten das Vorgängersymbol
     */
    override func KnotenZeichnen(xmin: Int, xmax: Int, y: Int, dy: Int, symbolVorgänger: KnotenSymbol?)
     {
         knotenSymbol.PositionSetzen(x: xmin + (xmax - xmin) / 2, y: y)
         // Eine Kante wird nur gezeichnet, wenn es einen Vorgänger gibt (d.h. dieser Knoten nicht die Wurzel ist)
         if symbolVorgänger != nil
         {
             if kantenSymbol == nil
             {
                 kantenSymbol = KantenSymbol(start: symbolVorgänger!, ende: knotenSymbol, gerichtet: false, gewicht: "", breite: 3, farbe: "schwarz")
             }
             else
             {
                 kantenSymbol!.DarstellungAktualisieren()
             }
         }
         rechterNachfolger.KnotenZeichnen(xmin: xmin + (xmax - xmin) / 2, xmax: xmax, y: y + dy, dy: dy, symbolVorgänger: knotenSymbol)
         linkerNachfolger.KnotenZeichnen(xmin: xmin, xmax: xmin + (xmax - xmin) / 2, y: y + dy, dy: dy, symbolVorgänger: knotenSymbol)
     }
}

/**
Die Klasse Binbaum ist die Grundstruktur eines geordneten Binärbaums
und hier zur Verwaltung von Usern eingesetzt wird.
Die Methoden die ein Objekt dieser Klasse anbietet, leiten den Aufruf an den
Wurzelknoten weiter, wenn der Baum nicht leer ist.
(Implementierung ohne Entwurfsmuster Kompositum)

- author: Peter Brichzin
- version: 24.5.24
*/
 class BinBaum
{
    /** Die Wurzel des Baums */
     private var wurzel: Baumelement

    /**
    Konstruktor für Objekte der Klasse Binbaum:
    Ein leerer Baum wird erzeugt.
    */
    init()
    {
        wurzel = Abschluss()
    }

    /**
    Fügt ein Datenelement sortiert in den geordneten Binärbaum ein,
    d.h. die Eigenschaft, dass der Baum geordnet ist bleibt bei jedem
    Einfügevorgang erhalten.
    - parameters:
        - datenNeu: neues Datenelement
    */
    func Einfügen(datenNeu: Wörterbucheintrag)
    {
        wurzel = wurzel.Einfügen(datenNeu: datenNeu)

        //Visualisierung des Baums
        BaumZeichnen()
    }

     /**
     Sucht ein Datenelement, das über seinen Schlüssel
     identifiziert wird, und gibt es aus.
     - parameters:
         - suchSchlüssel: Schlüssel, nach dem gesucht wird
     - returns: gesuchtes Datenelement bzw. nil, falls die Suche erfolglos ist
     */
     func Suchen(suchSchlüssel: String) -> Wörterbucheintrag?
     {
         return wurzel.Suchen(suchSchlüssel: suchSchlüssel)
     }

     /**
     Überprüft, ob ein Datenelement mit dem eingegebenen Schlüssel (in Form einer
     Zeichenkette) vorhanden ist.
     - parameters:
         - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
     - returns: true, im Erfolgsfall, false sonst.
     */
     func IstVorhanden(suchSchlüssel: String) -> Bool
     {
         return wurzel.IstVorhanden(suchSchlüssel: suchSchlüssel)
     }

     /**
     Berechnet die Höhe des Baums und gibt diesen Wert zurück
     - returns: Höhe des Baums
     */
     func HöheGeben() -> Int
     {
         return wurzel.HöheGeben(aktTiefe: 0)
     }

     /**
     Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schlüssel suchSchlüssel gespeichert ist.
     - parameters:
         - suchSchlüssel: Schlüssel, nach dem gesucht wird als Zeichenkette
     - returns: Tiefe als positive Zahl, bzw. -1 falls der suchSchlüssel nicht vorhanden ist.
     */
     func TiefeGeben(suchSchlüssel: String) -> Int
     {
         return wurzel.TiefeGeben(suchSchlüssel: suchSchlüssel, tiefe: 0)
     }


    /**
    Zeichnet den Baum
    */
    func BaumZeichnen()
    {
         let höhenschritt = Int(anzeigeHöhe) / (HöheGeben()+2) //+2 statt +1, damit unten ein Rand bleibt
         let breite = Int(anzeigebreite)

         let maxAnzahlKnotenUntersteEbene = Int(pow(2.0, Double(HöheGeben())))
         let dx = breite / (maxAnzahlKnotenUntersteEbene+1) //+1, damit rechts und links ein Rand bleibt

        wurzel.KnotenZeichnen(xmin: 0, xmax: breite,  y: 25, dy: höhenschritt, symbolVorgänger: nil) // y= 20 für die 0te Ebene, damit der Knoten sichtbar ist
     }
}

let woerterb1 = Wörterbucheintrag(wortNeu: "clip", bedeutungNeu: "Clip, Spange, Klammer")
let woerterb2 = Wörterbucheintrag(wortNeu: "cat", bedeutungNeu: "Katze")
let woerterb3 = Wörterbucheintrag(wortNeu: "cup", bedeutungNeu: "Tasse, Becher")
let woerterb4 = Wörterbucheintrag(wortNeu: "cut", bedeutungNeu: "Schnitt, Öffnung, kürzen")
let woerterb5 = Wörterbucheintrag(wortNeu: "cube", bedeutungNeu: "Würfel")
let woerterb6 = Wörterbucheintrag(wortNeu: "cave", bedeutungNeu: "Höhle")
let binBaum1 = BinBaum()
binBaum1.Einfügen(datenNeu: woerterb1)
binBaum1.Einfügen(datenNeu: woerterb2)
binBaum1.Einfügen(datenNeu: woerterb3)
binBaum1.Einfügen(datenNeu: woerterb4)
let woerterb7 = Wörterbucheintrag(wortNeu: "crab", bedeutungNeu: "Krabbe")

